# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tempest.api.identity import base
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from keystone_tempest_plugin import clients
from keystone_tempest_plugin.tests.rbac.v3 import base as rbac_base


class IdentityV3RbacLimitTests(rbac_base.IdentityV3RbacBaseTests,
                               metaclass=abc.ABCMeta):

    @classmethod
    def setup_clients(cls):
        super(IdentityV3RbacLimitTests, cls).setup_clients()
        cls.persona = getattr(cls, 'os_%s' % cls.credentials[0])
        persona_mgr = clients.Manager(cls.persona.credentials)
        cls.client = persona_mgr.limits_client
        cls.admin_client = cls.os_system_admin
        admin_mgr = clients.Manager(cls.admin_client.credentials)
        cls.admin_reglim_client = admin_mgr.registered_limits_client
        cls.admin_limits_client = admin_mgr.limits_client

    @classmethod
    def resource_setup(cls):
        cls.region_id = cls.admin_client.regions_client.create_region(
            region_id=data_utils.rand_name())['region']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.regions_client.delete_region,
            cls.region_id)
        svc_client = cls.admin_client.identity_services_v3_client
        cls.service_id = svc_client.create_service(
            type=data_utils.rand_name())['service']['id']
        cls.addClassResourceCleanup(svc_client.delete_service, cls.service_id)
        cls.own_domain = cls.persona.credentials.domain_id
        cls.other_domain = cls.admin_client.domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.domains_client.delete_domain, cls.other_domain)
        cls.addClassResourceCleanup(
            cls.admin_client.domains_client.update_domain,
            domain_id=cls.other_domain,
            enabled=False)
        cls.own_project = cls.persona.credentials.project_id
        # if project-scoped, use existing project
        # else create project in domain
        if not cls.own_project:
            cls.own_project = cls.admin_client.projects_client.create_project(
                name=data_utils.rand_name(),
                domain_id=cls.own_domain)['project']['id']
            cls.addClassResourceCleanup(
                cls.admin_client.projects_client.delete_project,
                cls.own_project)
        cls.other_project = cls.admin_client.projects_client.create_project(
            name=data_utils.rand_name(),
            domain_id=cls.other_domain)['project']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.projects_client.delete_project, cls.other_project)
        cls.reg_limit = cls.admin_reglim_client.create_registered_limits(
            payload=[{
                "service_id": cls.service_id,
                "region_id": cls.region_id,
                "resource_name": data_utils.rand_name(),
                "default_limit": 5
            }])['registered_limits'][0]
        cls.addClassResourceCleanup(
            cls.admin_reglim_client.delete_registered_limit,
            registered_limit_id=cls.reg_limit['id'])

    def limits(self, project_id=None):
        return [
            {
                "service_id": self.service_id,
                "region_id": self.region_id,
                "project_id": project_id or self.other_project,
                "resource_name": self.reg_limit['resource_name'],
                "resource_limit": 10
            }
        ]

    @abc.abstractmethod
    def test_identity_get_limit_model(self):
        """Test identity:get_limit_model policy.

        This test must check:
          * whether the persona can get the limit model
        """
        pass

    @abc.abstractmethod
    def test_identity_create_limits(self):
        """Test identity:create_limits policy.

        This test must check:
          * whether the persona can create a project limit for a project in any
            domain
          * whether the persona can create a project limit for a project in own
            domain
          * whether the persona can create a project limit for own project
        """
        pass

    @abc.abstractmethod
    def test_identity_list_limits(self):
        """Test identity:list_limits policy.

        This test must check:
          * whether the persona can list limits for any project
          * whether the persona can list limits for projects in own domain
          * whether the persona can list limits for own project
        """
        pass

    @abc.abstractmethod
    def test_identity_get_limit(self):
        """Test identity:get_limit policy.

        This test must check:
          * whether the persona can get a project limit for a project in any
            domain
          * whether the persona can get a project limit for a project in own
            domain
          * whether the persona can get a project limit for own project
        """
        pass

    @abc.abstractmethod
    def test_identity_update_limit(self):
        """Test identity:update_limit policy.

        This test must check:
          * whether the persona can update a project limit for a project in any
            domain
          * whether the persona can update a project limit for a project in own
            domain
          * whether the persona can update a project limit for own project
        """
        pass

    @abc.abstractmethod
    def test_identity_delete_limit(self):
        """Test identity:delete_limit policy.

        This test must check:
          * whether the persona can delete a project limit for a project limit
            in any domain
          * whether the persona can delete a project limit for a project limit
            in own domain
          * whether the persona can delete a project limit for a project limit
            own project
        """
        pass


class SystemAdminTests(IdentityV3RbacLimitTests, base.BaseIdentityTest):

    credentials = ['system_admin']

    def test_identity_get_limit_model(self):
        self.do_request('limits_model')

    def test_identity_create_limits(self):
        resp = self.do_request('create_limits',
                               expected_status=201,
                               payload=self.limits())
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=resp['limits'][0]['id'])

    def test_identity_list_limits(self):
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits(project_id=self.persona.credentials.project_id)
        )['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        resp = self.do_request('list_limits')
        self.assertIn(
            reg_limit_id, [rl['id'] for rl in resp['limits']])

    def test_identity_get_limit(self):
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        self.do_request('show_limit',
                        limit_id=reg_limit_id)

    def test_identity_update_limit(self):
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        updated = {'description': data_utils.arbitrary_string()}
        self.do_request('update_limit',
                        limit_id=reg_limit_id,
                        limit=updated)

    def test_identity_delete_limit(self):
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.do_request('delete_limit',
                        expected_status=204,
                        limit_id=reg_limit_id)


class SystemMemberTests(SystemAdminTests):

    credentials = ['system_member', 'system_admin']

    def test_identity_create_limits(self):
        self.do_request('create_limits',
                        expected_status=exceptions.Forbidden,
                        payload=self.limits())

    def test_identity_update_limit(self):
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        updated = {'description': data_utils.arbitrary_string()}
        self.do_request('update_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_id,
                        limit=updated)

    def test_identity_delete_limit(self):
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        self.do_request('delete_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_id)


class SystemReaderTests(SystemMemberTests):

    credentials = ['system_reader', 'system_admin']


class DomainAdminTests(IdentityV3RbacLimitTests, base.BaseIdentityTest):

    credentials = ['domain_admin', 'system_admin']

    def test_identity_get_limit_model(self):
        self.do_request('limits_model')

    def test_identity_create_limits(self):
        # cannot create limit in arbitrary project
        self.do_request('create_limits',
                        expected_status=exceptions.Forbidden,
                        payload=self.limits())
        # cannot create limit in project in own domain
        self.do_request('create_limits',
                        expected_status=exceptions.Forbidden,
                        payload=self.limits(project_id=self.own_project))

    def test_identity_list_limits(self):
        # random project
        reg_limit_1 = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_1)
        # project in own domain
        reg_limit_2 = self.admin_limits_client.create_limits(
            payload=self.limits(project_id=self.own_project)
        )['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_2)
        resp = self.do_request('list_limits')
        # admin should see limit for other project
        self.assertIn(
            reg_limit_1, [rl['id'] for rl in resp['limits']])
        # should see limit for project in own domain
        self.assertIn(
            reg_limit_2, [rl['id'] for rl in resp['limits']])

    def test_identity_get_limit(self):
        # random project
        reg_limit_1 = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_1)
        self.do_request('show_limit',
                        limit_id=reg_limit_1)

    def test_identity_update_limit(self):
        # cannot update limit for arbitrary project
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        updated = {'description': data_utils.arbitrary_string()}
        self.do_request('update_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_id,
                        limit=updated)
        # cannot update limit for project in own domain
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits(project_id=self.own_project)
        )['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        updated = {'description': data_utils.arbitrary_string()}
        self.do_request('update_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_id,
                        limit=updated)

    def test_identity_delete_limit(self):
        # cannot delete limit for arbitrary project
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        self.do_request('delete_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_id)

        # cannot delete limit for project in own domain
        reg_limit_id = self.admin_limits_client.create_limits(
            payload=self.limits(project_id=self.own_project)
        )['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_id)
        self.do_request('delete_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_id)


class DomainManagerTests(DomainAdminTests):

    credentials = ['domain_manager', 'system_admin']

    def test_identity_list_limits(self):
        # random project
        reg_limit_1 = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_1)
        # project in own domain
        reg_limit_2 = self.admin_limits_client.create_limits(
            payload=self.limits(project_id=self.own_project)
        )['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_2)
        resp = self.do_request('list_limits')
        # should not see limit for other project
        self.assertNotIn(
            reg_limit_1, [rl['id'] for rl in resp['limits']])
        # should see limit for project in own domain
        self.assertIn(
            reg_limit_2, [rl['id'] for rl in resp['limits']])

    def test_identity_get_limit(self):
        # random project
        reg_limit_1 = self.admin_limits_client.create_limits(
            payload=self.limits())['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_1)
        # project in own domain
        reg_limit_2 = self.admin_limits_client.create_limits(
            payload=self.limits(project_id=self.own_project)
        )['limits'][0]['id']
        self.addCleanup(
            self.admin_limits_client.delete_limit,
            limit_id=reg_limit_2)
        # cannot get limit for other project
        self.do_request('show_limit',
                        expected_status=exceptions.Forbidden,
                        limit_id=reg_limit_1)
        # can get limit for project in own domain
        self.do_request('show_limit',
                        limit_id=reg_limit_2)


class DomainMemberTests(DomainManagerTests):

    credentials = ['domain_member', 'system_admin']


class DomainReaderTests(DomainMemberTests):

    credentials = ['domain_reader', 'system_admin']


class ProjectAdminTests(SystemAdminTests):

    credentials = ['project_admin', 'system_admin']


class ProjectManagerTests(DomainReaderTests):

    credentials = ['project_manager', 'system_admin']


class ProjectMemberTests(ProjectManagerTests):

    credentials = ['project_member', 'system_admin']


class ProjectReaderTests(ProjectMemberTests):

    credentials = ['project_reader', 'system_admin']



# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tempest.api.identity import base
from tempest import clients
from tempest.lib import auth
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from keystone_tempest_plugin.tests.rbac.v3 import base as rbac_base


class IdentityV3RbacAccessRuleTest(rbac_base.IdentityV3RbacBaseTests,
                                   metaclass=abc.ABCMeta):

    identity_version = 'v3'

    @classmethod
    def setup_clients(cls):
        super(IdentityV3RbacAccessRuleTest, cls).setup_clients()
        cls.persona = getattr(cls, 'os_%s' % cls.credentials[0])
        cls.client = cls.persona.access_rules_client
        cls.admin_client = cls.os_system_admin

    def user(self):
        user = {}
        name = data_utils.rand_name('user')
        user['name'] = name
        user['password'] = data_utils.rand_password()
        return user

    def app_cred(self):
        app_cred = {}
        app_cred['name'] = data_utils.rand_name('app_cred')
        app_cred['access_rules'] = [
            {
                'path': '/servers',
                'method': 'GET',
                'service': 'compute',
            }
        ]
        return app_cred

    @classmethod
    def setup_user_client(cls, domain_id=None):
        """Set up project user with its own client.

        This is to enable the project user to create its own app cred.

        Returns a client object and the user's ID.
        """
        user_dict = {
            'name': data_utils.rand_name('user'),
            'password': data_utils.rand_password(),
        }
        if domain_id:
            user_dict['domain_id'] = domain_id
        user_id = cls.admin_client.users_v3_client.create_user(
            **user_dict)['user']['id']

        def try_delete_user():
            # delete user if not deleted by domain deletion
            try:
                cls.admin_client.users_v3_client.delete_user(user_id)
            except exceptions.NotFound:
                pass

        cls.addClassResourceCleanup(try_delete_user)
        project_id = cls.admin_client.projects_client.create_project(
            data_utils.rand_name())['project']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.projects_client.delete_project, project_id)
        member_role_id = cls.admin_client.roles_v3_client.list_roles(
            name='member')['roles'][0]['id']
        cls.admin_client.roles_v3_client.create_user_role_on_project(
            project_id, user_id, member_role_id)
        creds = auth.KeystoneV3Credentials(
            user_id=user_id,
            password=user_dict['password'],
            project_id=project_id)
        auth_provider = clients.get_auth_provider(creds)
        creds = auth_provider.fill_credentials()
        client = clients.Manager(credentials=creds)
        return client, user_id

    @abc.abstractmethod
    def test_identity_get_access_rule(self):
        """Test identity:get_access_rule policy

        This test must check:
          * whether the persona can retrieve an access rule they own
          * whether the persona can retrieve an access rule they do not own
          * whether the persona can retrieve an access rule that does not exist
          * whether the persona can retrieve an access rule for a user in their
            own domain (if applicable)
          * whether the persona can retrieve an access rule for a user in
            another domain (if applicable)
        """
        pass

    @abc.abstractmethod
    def test_identity_list_access_rules(self):
        """Test identity:list_access_rules policy

        This test must check:
          * whether the persona can list their own access rules
          * whether the persona can list the access rules for another user
          * whether the persona can list the access rules for a user in their
            own domain
          * whether the persona can list the access rules for a user in another
            domain
        """
        pass

    @abc.abstractmethod
    def test_identity_delete_access_rule(self):
        """Test identity:delete_access_rule policy.

        This test must check
          * whether the persona can delete an access rule they own
          * whether the persona can delete an access rule for an arbitrary user
          * whether the persona can delete an access rule that does not exist
          * whether the persona can delete an access rule for a user in another
            domain (if applicable)
          * whether the persona can delete an access rule for a user in their
            own domain (if applicable)
          * whether the persona can delete an access rule that does not exist
        """
        pass


class SystemAdminTests(IdentityV3RbacAccessRuleTest, base.BaseIdentityTest):

    credentials = ['system_admin']

    @classmethod
    def setup_clients(cls):
        super(SystemAdminTests, cls).setup_clients()
        cls.test_user_client, cls.test_user_id = cls.setup_user_client()

    def setUp(self):
        # create app cred for other user
        super(SystemAdminTests, self).setUp()
        app_cred_client = self.test_user_client.application_credentials_client
        app_cred = app_cred_client.create_application_credential(
            user_id=self.test_user_id, **self.app_cred()
        )['application_credential']
        self.app_cred_id = app_cred['id']
        self.access_rule_id = app_cred['access_rules'][0]['id']

        def try_delete_app_cred(id):
            app_cred_client = self.admin_client.application_credentials_client
            try:
                app_cred_client.delete_application_credential(
                    user_id=self.test_user_id,
                    application_credential_id=id)
            except exceptions.NotFound:
                pass

        def try_delete_access_rule(id):
            try:
                self.admin_client.access_rules_client.delete_access_rule(
                    user_id=self.test_user_id,
                    access_rule_id=id)
            except exceptions.NotFound:
                pass
        self.addCleanup(try_delete_access_rule, self.access_rule_id)
        self.addCleanup(try_delete_app_cred, self.app_cred_id)

    def test_identity_get_access_rule(self):
        # system admin cannot create app creds and therefore cannot create
        # access rules, so skip retrieval of own access rule

        # retrieve other user's access rules
        self.do_request(
            'show_access_rule',
            user_id=self.test_user_id, access_rule_id=self.access_rule_id)

        # retrieving a non-existent access rule should return a 404
        self.do_request(
            'show_access_rule', expected_status=exceptions.NotFound,
            user_id=self.test_user_id,
            access_rule_id=data_utils.rand_uuid_hex())

    def test_identity_list_access_rules(self):
        # system admin cannot create app creds and therefore cannot create
        # access rules, so skip listing of own access rule

        # list other user's access rules
        self.do_request('list_access_rules', user_id=self.test_user_id)

    def test_identity_delete_access_rule(self):
        # system admin cannot create app creds and therefore cannot create
        # access rules, so skip deletion of own access rule

        # delete other user's access rules
        app_cred_client = self.admin_client.application_credentials_client
        app_cred_client.delete_application_credential(
            user_id=self.test_user_id,
            application_credential_id=self.app_cred_id)
        self.do_request(
            'delete_access_rule', expected_status=204,
            user_id=self.test_user_id, access_rule_id=self.access_rule_id)

        # deleting a non-existent access rule should return a 404
        self.do_request(
            'delete_access_rule', expected_status=exceptions.NotFound,
            user_id=self.test_user_id,
            access_rule_id=data_utils.rand_uuid_hex())


class SystemMemberTests(SystemAdminTests):

    credentials = ['system_member', 'system_admin']

    def test_identity_delete_access_rule(self):
        app_cred_client = self.admin_client.application_credentials_client
        app_cred_client.delete_application_credential(
            user_id=self.test_user_id,
            application_credential_id=self.app_cred_id)
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_id, access_rule_id=self.access_rule_id)

        # retrieving a non-existent access rule should return a 404
        self.do_request(
            'show_access_rule', expected_status=exceptions.NotFound,
            user_id=self.test_user_id,
            access_rule_id=data_utils.rand_uuid_hex())


class SystemReaderTests(SystemMemberTests):

    credentials = ['system_reader', 'system_admin']


class DomainAdminTests(IdentityV3RbacAccessRuleTest, base.BaseIdentityTest):

    # Domain admins cannot create their own app creds (app creds can only be
    # scoped to projects) and domain admins have no special privileges over the
    # app creds own by users in their domains.

    credentials = ['domain_admin', 'system_admin']

    @classmethod
    def setup_clients(cls):
        super(DomainAdminTests, cls).setup_clients()
        own_domain_id = cls.persona.credentials.domain_id
        cls.test_client_1, cls.test_user_1 = cls.setup_user_client(
            domain_id=own_domain_id)

    def setUp(self):
        super(DomainAdminTests, self).setUp()
        self.other_domain_id = self.admin_client.domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_client.domains_client.delete_domain,
                        self.other_domain_id)
        self.addCleanup(self.admin_client.domains_client.update_domain,
                        domain_id=self.other_domain_id, enabled=False)
        self.test_client_2, self.test_user_2 = self.setup_user_client(
            domain_id=self.other_domain_id)
        client = self.test_client_1.application_credentials_client
        app_cred_1 = client.create_application_credential(
            user_id=self.test_user_1, **self.app_cred()
        )['application_credential']
        self.access_rule_1 = app_cred_1['access_rules'][0]['id']
        self.addCleanup(
            self.test_client_1.access_rules_client.delete_access_rule,
            self.test_user_1,
            self.access_rule_1)
        self.addCleanup(
            client.delete_application_credential,
            self.test_user_1,
            app_cred_1['id'])
        client = self.test_client_2.application_credentials_client
        app_cred_2 = client.create_application_credential(
            user_id=self.test_user_2, **self.app_cred()
        )['application_credential']
        self.access_rule_2 = app_cred_2['access_rules'][0]['id']
        self.addCleanup(
            self.test_client_2.access_rules_client.delete_access_rule,
            self.test_user_2,
            self.access_rule_2)
        self.addCleanup(
            client.delete_application_credential,
            self.test_user_2,
            app_cred_2['id'])

    def test_identity_get_access_rule(self):
        # accessing access rules should be forbidden no matter whether the
        # owner is in the domain or outside of it

        # retrieve access rule from user in own domain
        self.do_request(
            'show_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_1, access_rule_id=self.access_rule_1)

        # retrieve access rule from user in other domain
        self.do_request(
            'show_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_2, access_rule_id=self.access_rule_2)

        # retrieving a non-existent access rule should return a 403
        self.do_request(
            'show_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_1,
            access_rule_id=data_utils.rand_uuid_hex())
        self.do_request(
            'show_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_2,
            access_rule_id=data_utils.rand_uuid_hex())

    def test_identity_list_access_rules(self):
        # listing access rules should be forbidden no matter whether the
        # owner is in the domain or outside of it
        self.do_request(
            'list_access_rules', expected_status=exceptions.Forbidden,
            user_id=self.test_user_1)
        self.do_request(
            'list_access_rules', expected_status=exceptions.Forbidden,
            user_id=self.test_user_2)

    def test_identity_delete_access_rule(self):
        # deleting access rules should be forbidden no matter whether the
        # owner is in the domain or outside of it

        # delete access rule from user in own domain
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_1, access_rule_id=self.access_rule_1)

        # delete access rule from user in other domain
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_2, access_rule_id=self.access_rule_2)

        # deleting a non-existent access rule should return a 403
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_1,
            access_rule_id=data_utils.rand_uuid_hex())
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_2,
            access_rule_id=data_utils.rand_uuid_hex())


class DomainManagerTests(DomainAdminTests):

    credentials = ['domain_manager', 'system_admin']


class DomainMemberTests(DomainManagerTests):

    credentials = ['domain_member', 'system_admin']


class DomainReaderTests(DomainAdminTests):

    credentials = ['domain_reader', 'system_admin']


class ProjectAdminTests(IdentityV3RbacAccessRuleTest, base.BaseIdentityTest):

    credentials = ['project_admin', 'system_admin']

    @classmethod
    def setup_clients(cls):
        super(ProjectAdminTests, cls).setup_clients()
        cls.test_user_client, cls.test_user_id = cls.setup_user_client()

    def setUp(self):
        super(ProjectAdminTests, self).setUp()
        app_cred_client = self.persona.application_credentials_client
        user_id = self.persona.credentials.user_id
        self.app_cred_1 = app_cred_client.create_application_credential(
            user_id, **self.app_cred())['application_credential']
        self.access_rule_1 = self.app_cred_1['access_rules'][0]['id']

        def try_delete_own_app_cred(id):
            app_cred_client = self.persona.application_credentials_client
            try:
                app_cred_client.delete_application_credential(
                    self.persona.credentials.user_id, id)
            except exceptions.NotFound:
                pass

        def try_delete_own_access_rule(id):
            try:
                self.persona.access_rules_client.delete_access_rule(
                    self.persona.credentials.user_id, id)
            except exceptions.NotFound:
                pass

        self.addCleanup(try_delete_own_access_rule, self.access_rule_1)
        self.addCleanup(try_delete_own_app_cred, self.app_cred_1['id'])

        app_cred_client = self.test_user_client.application_credentials_client
        self.app_cred_2 = app_cred_client.create_application_credential(
            self.test_user_id, **self.app_cred())['application_credential']
        self.access_rule_2 = self.app_cred_2['access_rules'][0]['id']
        self.addCleanup(
            self.test_user_client.access_rules_client.delete_access_rule,
            self.test_user_id, self.access_rule_2)
        self.addCleanup(
            app_cred_client.delete_application_credential,
            self.test_user_id, self.app_cred_2['id'])

    def test_identity_get_access_rule(self):
        # should be able to access own credential
        self.do_request(
            'show_access_rule',
            user_id=self.persona.credentials.user_id,
            access_rule_id=self.access_rule_1)

        # retrieving non-existent access rule for self should return 404
        self.do_request(
            'show_access_rule', expected_status=exceptions.NotFound,
            user_id=self.persona.credentials.user_id,
            access_rule_id=data_utils.rand_uuid_hex())

        # should not be able to access another user's credential
        self.do_request(
            'show_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_id, access_rule_id=self.access_rule_2)

        # retrieving non-existent access rule for other user should return 403
        self.do_request(
            'show_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_id,
            access_rule_id=data_utils.rand_uuid_hex())

    def test_identity_list_access_rules(self):
        # should be able to list own credentials
        self.do_request(
            'list_access_rules', user_id=self.persona.credentials.user_id)

        # should not be able to list another user's credentials
        self.do_request(
            'list_access_rules', expected_status=exceptions.Forbidden,
            user_id=self.test_user_id)

    def test_identity_delete_access_rule(self):
        # should be able to delete own credential
        app_cred_client = self.persona.application_credentials_client
        app_cred_client.delete_application_credential(
            user_id=self.persona.credentials.user_id,
            application_credential_id=self.app_cred_1['id'])
        self.do_request(
            'delete_access_rule', expected_status=204,
            user_id=self.persona.credentials.user_id,
            access_rule_id=self.access_rule_1)

        # deleting non-existent access rule for self should return 404
        self.do_request(
            'delete_access_rule', expected_status=exceptions.NotFound,
            user_id=self.persona.credentials.user_id,
            access_rule_id=data_utils.rand_uuid_hex())

        # should not be able to delete another user's credential
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_id, access_rule_id=self.access_rule_2)

        # deleting non-existent access rule for other user should return 403
        self.do_request(
            'delete_access_rule', expected_status=exceptions.Forbidden,
            user_id=self.test_user_id,
            access_rule_id=data_utils.rand_uuid_hex())


class ProjectManagerTests(ProjectAdminTests):

    credentials = ['project_manager', 'system_admin']


class ProjectMemberTests(ProjectManagerTests):

    credentials = ['project_member', 'system_admin']


class ProjectReaderTests(ProjectAdminTests):

    credentials = ['project_reader', 'system_admin']



# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tempest.api.identity import base
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from keystone_tempest_plugin.tests.rbac.v3 import base as rbac_base


class IdentityV3RbacDomainTests(rbac_base.IdentityV3RbacBaseTests,
                                metaclass=abc.ABCMeta):

    @classmethod
    def setup_clients(cls):
        super(IdentityV3RbacDomainTests, cls).setup_clients()
        cls.persona = getattr(cls, 'os_%s' % cls.credentials[0])
        cls.client = cls.persona.domains_client
        admin_client = cls.os_system_admin
        cls.admin_domains_client = admin_client.domains_client
        cls.own_domain = cls.persona.credentials.domain_id

    @abc.abstractmethod
    def test_identity_create_domain(self):
        """Test identity:create_domain policy.

        This test must check:
          * whether the persona can create a domain
        """
        pass

    @abc.abstractmethod
    def test_identity_get_domain(self):
        """Test identity:get_domain policy.

        This test must check:
          * whether the persona can get a domain
          * whether the persona can get a domain that does not exist
        """
        pass

    @abc.abstractmethod
    def test_identity_list_domains(self):
        """Test identity:list_domains policy.

        This test must check:
          * whether the persona can list all domains
        """
        pass

    @abc.abstractmethod
    def test_identity_update_domain(self):
        """Test identity:update_domain policy.

        This test must check:
          * whether the persona can update a domain
          * whether the persona can update a domain that does not exist
        """
        pass

    @abc.abstractmethod
    def test_identity_delete_domain(self):
        """Test identity:delete_domain policy.

        This test must check
          * whether the persona can delete a domain
          * whether the persona can delete a domain that does not exist
        """
        pass


class SystemAdminTests(IdentityV3RbacDomainTests, base.BaseIdentityTest):

    credentials = ['system_admin']

    def test_identity_create_domain(self):
        domain_id = self.do_request(
            'create_domain', expected_status=201, name=data_utils.rand_name()
        )['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)

    def test_identity_get_domain(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        self.do_request('show_domain', domain_id=domain_id)
        # user gets a 404 for nonexistent domain
        self.do_request('show_domain', expected_status=exceptions.NotFound,
                        domain_id=data_utils.rand_uuid_hex())

    def test_identity_list_domains(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        resp = self.do_request('list_domains')
        self.assertIn(domain_id, [d['id'] for d in resp['domains']])

    def test_identity_update_domain(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        self.do_request('update_domain',
                        domain_id=domain_id,
                        description=data_utils.arbitrary_string())
        # user gets a 404 for nonexistent domain
        self.do_request('update_domain', expected_status=exceptions.NotFound,
                        domain_id=data_utils.rand_uuid_hex())

    def test_identity_delete_domain(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.do_request('update_domain',
                        domain_id=domain_id,
                        enabled=False)
        self.do_request('delete_domain', expected_status=204,
                        domain_id=domain_id)


class SystemMemberTests(SystemAdminTests, base.BaseIdentityTest):

    credentials = ['system_member', 'system_admin']

    def test_identity_create_domain(self):
        self.do_request('create_domain', expected_status=exceptions.Forbidden,
                        name=data_utils.rand_name())

    def test_identity_update_domain(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        self.do_request('update_domain', expected_status=exceptions.Forbidden,
                        domain_id=domain_id,
                        description=data_utils.arbitrary_string())
        # user gets a 404 for nonexistent domain
        self.do_request('update_domain', expected_status=exceptions.NotFound,
                        domain_id=data_utils.rand_uuid_hex())

    def test_identity_delete_domain(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        self.do_request('delete_domain', expected_status=exceptions.Forbidden,
                        domain_id=domain_id)


class SystemReaderTests(SystemMemberTests):

    credentials = ['system_reader', 'system_admin']


class DomainAdminTests(SystemReaderTests, base.BaseIdentityTest):

    credentials = ['domain_admin', 'system_admin']

    def test_identity_list_domains(self):
        domain_id = self.persona.credentials.domain_id
        other_domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain,
                        other_domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=other_domain_id, enabled=False)
        resp = self.do_request('list_domains')
        self.assertIn(domain_id, [d['id'] for d in resp['domains']])
        self.assertNotIn(other_domain_id, [d['id'] for d in resp['domains']])


class DomainManagerTests(DomainAdminTests, base.BaseIdentityTest):

    credentials = ['domain_manager', 'system_admin']

    def test_identity_get_domain(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        # user can get own domain
        self.do_request('show_domain', domain_id=self.own_domain)
        # user gets a 403 for foreign domain
        self.do_request('show_domain', expected_status=exceptions.Forbidden,
                        domain_id=domain_id)
        # user gets a 403 for nonexistent domain
        self.do_request('show_domain', expected_status=exceptions.Forbidden,
                        domain_id=data_utils.rand_uuid_hex())


class DomainMemberTests(DomainManagerTests):

    credentials = ['domain_member', 'system_admin']


class DomainReaderTests(DomainMemberTests):

    credentials = ['domain_reader', 'system_admin']


class ProjectAdminTests(SystemAdminTests):

    credentials = ['project_admin', 'system_admin']


class ProjectManagerTests(DomainReaderTests):

    credentials = ['project_manager', 'system_admin']

    def test_identity_list_domains(self):
        domain_id = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']['id']
        self.addCleanup(self.admin_domains_client.delete_domain, domain_id)
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=domain_id, enabled=False)
        self.do_request('list_domains',
                        expected_status=exceptions.Forbidden)


class ProjectMemberTests(ProjectManagerTests):

    credentials = ['project_member', 'system_admin']


class ProjectReaderTests(ProjectMemberTests):

    credentials = ['project_reader', 'system_admin']



# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tempest.api.identity import base
from tempest import clients
from tempest.lib import auth
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from keystone_tempest_plugin.tests.rbac.v3 import base as rbac_base


class IdentityV3RbacTrustTest(rbac_base.IdentityV3RbacBaseTests,
                              metaclass=abc.ABCMeta):

    @classmethod
    def setup_clients(cls):
        super(IdentityV3RbacTrustTest, cls).setup_clients()
        cls.persona = getattr(cls, 'os_%s' % cls.credentials[0])
        cls.client = cls.persona.trusts_client
        cls.admin_client = cls.os_system_admin
        cls.admin_trusts_client = cls.admin_client.trusts_client

    @classmethod
    def resource_setup(cls):
        trustor_user = {
            'name': data_utils.rand_name('user'),
            'password': data_utils.rand_password(),
        }
        cls.trustor = cls.admin_client.users_v3_client.create_user(
            **trustor_user)['user']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.users_v3_client.delete_user,
            user_id=cls.trustor)
        cls.trustee = cls.admin_client.users_v3_client.create_user(
            name=data_utils.rand_name())['user']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.users_v3_client.delete_user,
            user_id=cls.trustee)
        cls.project = cls.admin_client.projects_client.create_project(
            name=data_utils.rand_name()
        )['project']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.projects_client.delete_project,
            project_id=cls.project)
        cls.roles = [
            {'id': cls.admin_client.roles_v3_client.create_role(
                name=data_utils.rand_name())['role']['id']}
        ]
        cls.addClassResourceCleanup(
            cls.admin_client.roles_v3_client.delete_role,
            role_id=cls.roles[0]['id'])
        cls.admin_client.roles_v3_client.create_user_role_on_project(
            project_id=cls.project,
            user_id=cls.trustor,
            role_id=cls.roles[0]['id']
        )
        creds = auth.KeystoneV3Credentials(
            user_id=cls.trustor,
            password=trustor_user['password'],
            project_id=cls.project)
        auth_provider = clients.get_auth_provider(creds)
        creds = auth_provider.fill_credentials()
        user_client = clients.Manager(credentials=creds)
        cls.user_trust_client = user_client.trusts_client

        cls.admin_role_id = cls.admin_client.roles_v3_client.list_roles(
            name='admin')['roles'][0]['id']
        cls.member_role_id = cls.admin_client.roles_v3_client.list_roles(
            name='member')['roles'][0]['id']
        cls.reader_role_id = cls.admin_client.roles_v3_client.list_roles(
            name='reader')['roles'][0]['id']

    def trust(self, trustor=None, trustee=None, project_id=None, roles=None):
        trust = {}
        trust['trustor_user_id'] = trustor or self.trustor
        trust['trustee_user_id'] = trustee or self.trustee
        trust['project_id'] = project_id or self.project
        trust['roles'] = roles or self.roles
        trust['impersonation'] = False
        return trust

    @abc.abstractmethod
    def test_identity_create_trust(self):
        """Test identity:create_trust policy.

        This test must check:
          * whether the persona can create a trust for themself
          * whether the persona can create a trust for another user
        """
        pass

    @abc.abstractmethod
    def test_identity_get_trust(self):
        """Test identity:get_trust policy.

        This test must check:
          * whether the persona can get a trust for which they are the trustor
          * whether the persona can get a trust for which they are the trustee
          * whether the persona can get a trust with which they are
            unaffiliated
        """
        pass

    @abc.abstractmethod
    def test_identity_list_trusts(self):
        """Test identity:list_trusts policy.

        This test must check:
          * whether the persona can list all trusts
        """
        pass

    @abc.abstractmethod
    def test_identity_list_trusts_for_trustor(self):
        """Test identity:list_trusts_for_trustor policy.

        This test must check:
          * whether the persona can list trusts by trustor for which they are
            the trustor
          * whether the persona can list trusts by trustor for which another
            user is trustor
        """
        pass

    @abc.abstractmethod
    def test_identity_list_trusts_for_trustee(self):
        """Test identity:list_trusts_for_trustee policy.

        This test must check:
          * whether the persona can list trusts by trustee for which they are
            the trustee
          * whether the persona can list trusts by trustee for which another
            user is trustee
        """
        pass

    @abc.abstractmethod
    def test_identity_list_roles_for_trust(self):
        """Test identity:list_roles_for_trust policy.

        This test must check:
          * whether the persona can list the roles of a trust for which they
            are the trustee
          * whether the persona can list the roles of a trust for which they
            are the trustor
          * whether the persona can list the roles of a trust with which they
            are unaffiliated
        """
        pass

    @abc.abstractmethod
    def test_identity_get_role_for_trust(self):
        """Test identity:get_role_for_trust policy.

        This test must check:
          * whether the persona can get a role of a trust for which they are
            the trustee
          * whether the persona can get a role of a trust for which they are
            the trustor
          * whether the persona can get a role of a trust with which they are
            unaffiliated
        """
        pass

    @abc.abstractmethod
    def test_identity_delete_trust(self):
        """Test identity:delete_trust policy.

        This test must check
          * whether the persona can delete a trust for which they are the
            trustor
          * whether the persona can delete a trust for which they are the
            trustee
          * whether the persona can delete a trust which which they are
            unaffiliated
        """
        pass


class SystemAdminTests(IdentityV3RbacTrustTest, base.BaseIdentityTest):

    credentials = ['system_admin']

    def test_identity_create_trust(self):
        # user cannot create trust for themself
        self.do_request('create_trust',
                        expected_status=exceptions.Forbidden,
                        **self.trust(trustor=self.persona.credentials.user_id))
        # user cannot create trust for another user
        self.do_request('create_trust',
                        expected_status=exceptions.Forbidden,
                        **self.trust())

    def test_identity_get_trust(self):
        # user cannot have their own trust
        # user can get trust for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('show_trust', trust_id=trust_id)

    def test_identity_list_trusts(self):
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        resp = self.do_request('list_trusts')
        self.assertIn(trust_id, [t['id'] for t in resp['trusts']])

    def test_identity_list_trusts_for_trustor(self):
        # user cannot have their own trust
        # user can list trusts for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trusts', trustor_user_id=self.trustor)

    def test_identity_list_trusts_for_trustee(self):
        # user cannot have their own trust
        # user can list trusts for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trusts', trustee_user_id=self.trustee)

    def test_identity_list_roles_for_trust(self):
        # user cannot have their own trust
        # user can list roles of trust for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        resp = self.do_request('list_trust_roles', trust_id=trust_id)
        self.assertIn(self.roles[0]['id'], [r['id'] for r in resp['roles']])

    def test_identity_get_role_for_trust(self):
        # user cannot have their own trust
        # user can get role of trust for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('show_trust_role',
                        trust_id=trust_id, role_id=self.roles[0]['id'])

    def test_identity_delete_trust(self):
        # user cannot have their own trust
        # user can delete a user's trust
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.do_request('delete_trust', expected_status=204, trust_id=trust_id)


class SystemMemberTests(SystemAdminTests):

    credentials = ['system_member', 'system_admin']

    def test_identity_delete_trust(self):
        # system user cannot have their own trust
        # user cannot delete another user's trust
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('delete_trust', expected_status=exceptions.Forbidden,
                        trust_id=trust_id)


class SystemReaderTests(SystemMemberTests):

    credentials = ['system_reader', 'system_admin']


class DomainAdminTests(SystemReaderTests, base.BaseIdentityTest):

    # Domain admins cannot create their own trusts (trusts can only be
    # scoped to projects) and domain admins have no special privileges over the
    # trusts own by users in their domains.

    credentials = ['domain_admin', 'system_admin']

    def test_identity_get_trust(self):
        # user cannot have their own trust
        # user can get trust for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('show_trust', expected_status=exceptions.Forbidden,
                        trust_id=trust_id)

    def test_identity_list_trusts(self):
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trusts',
                        expected_status=exceptions.Forbidden)

    def test_identity_list_trusts_for_trustor(self):
        # user cannot have their own trust
        # user can list trusts for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trusts', expected_status=exceptions.Forbidden,
                        trustor_user_id=self.trustor)

    def test_identity_list_trusts_for_trustee(self):
        # user cannot have their own trust
        # user can list trusts for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trusts', expected_status=exceptions.Forbidden,
                        trustee_user_id=self.trustee)

    def test_identity_list_roles_for_trust(self):
        # user cannot have their own trust
        # user can list roles of trust for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trust_roles',
                        expected_status=exceptions.Forbidden,
                        trust_id=trust_id)

    def test_identity_get_role_for_trust(self):
        # user cannot have their own trust
        # user can get role of trust for other user
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('show_trust_role',
                        expected_status=exceptions.Forbidden,
                        trust_id=trust_id, role_id=self.roles[0]['id'])


class DomainManagerTests(DomainAdminTests):

    credentials = ['domain_manager', 'system_admin']


class DomainMemberTests(DomainManagerTests):

    credentials = ['domain_member', 'system_admin']


class DomainReaderTests(DomainAdminTests):

    credentials = ['domain_reader', 'system_admin']


class ProjectAdminTests(SystemAdminTests):

    credentials = ['project_admin', 'system_admin']

    def setUp(self):
        super(ProjectAdminTests, self).setUp()
        self.role_id = self.admin_role_id

    def test_identity_create_trust(self):
        # user can create a trust for their own project
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        resp = self.do_request(
            'create_trust',
            expected_status=201,
            **self.trust(
                trustor=trustor_user_id,
                project_id=project_id,
                roles=[{'id': self.role_id}])
        )['trust']
        self.addCleanup(self.client.delete_trust, resp['id'])

        # user cannot create trust with another user as trustor
        self.do_request(
            'create_trust',
            expected_status=exceptions.Forbidden,
            **self.trust())


class ProjectManagerTests(IdentityV3RbacTrustTest, base.BaseIdentityTest):

    credentials = ['project_manager', 'system_admin']

    def setUp(self):
        super(ProjectManagerTests, self).setUp()
        self.role_id = self.member_role_id

    def test_identity_create_trust(self):
        # user can create a trust for their own project
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        resp = self.do_request(
            'create_trust',
            expected_status=201,
            **self.trust(
                trustor=trustor_user_id,
                project_id=project_id,
                roles=[{'id': self.role_id}])
        )['trust']
        self.addCleanup(self.client.delete_trust, resp['id'])

        # user cannot create trust with another user as trustor
        self.do_request(
            'create_trust',
            expected_status=exceptions.Forbidden,
            **self.trust())

    def test_identity_get_trust(self):
        # user can get a trust for which they are trustor
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        trust_id = self.client.create_trust(
            **self.trust(trustor=trustor_user_id,
                         project_id=project_id,
                         roles=[{'id': self.role_id}]))['trust']['id']
        self.addCleanup(self.client.delete_trust, trust_id=trust_id)
        self.do_request('show_trust', trust_id=trust_id)

        # user can get a trust for which they are trustee
        trustee_user_id = self.persona.credentials.user_id
        trust_id = self.user_trust_client.create_trust(
            **self.trust(trustee=trustee_user_id))['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('show_trust', trust_id=trust_id)

        # user cannot get a trust with which they are unaffiliated
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('show_trust', expected_status=exceptions.Forbidden,
                        trust_id=trust_id)

    def test_identity_list_trusts(self):
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.admin_trusts_client.delete_trust,
                        trust_id=trust_id)
        self.do_request('list_trusts',
                        expected_status=exceptions.Forbidden)

    def test_identity_list_trusts_for_trustor(self):
        # user can list their own trusts
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        trust_id = self.client.create_trust(
            **self.trust(trustor=trustor_user_id,
                         project_id=project_id,
                         roles=[{'id': self.role_id}]))['trust']['id']
        self.addCleanup(self.client.delete_trust, trust_id=trust_id)
        self.do_request('list_trusts', trustor_user_id=trustor_user_id)

        # user cannot list another user's trusts
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('list_trusts', expected_status=exceptions.Forbidden,
                        trustor_user_id=self.trustor)

    def test_identity_list_trusts_for_trustee(self):
        # user can list their own trusts
        trustee_user_id = self.persona.credentials.user_id
        trust_id = self.user_trust_client.create_trust(
            **self.trust(trustee=trustee_user_id))['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('list_trusts', trustee_user_id=trustee_user_id)

        # user cannot list another user's trusts
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('list_trusts', expected_status=exceptions.Forbidden,
                        trustee_user_id=self.trustee)

    def test_identity_list_roles_for_trust(self):
        # user can list roles for trust for which they are trustor
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        trust_id = self.client.create_trust(
            **self.trust(trustor=trustor_user_id,
                         project_id=project_id,
                         roles=[{'id': self.role_id}]))['trust']['id']
        self.addCleanup(self.client.delete_trust, trust_id=trust_id)
        self.do_request('list_trust_roles', trust_id=trust_id)

        # user can list roles for trust for which they are trustee
        trustee_user_id = self.persona.credentials.user_id
        trust_id = self.user_trust_client.create_trust(
            **self.trust(trustee=trustee_user_id))['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('list_trust_roles', trust_id=trust_id)

        # user cannot list roles for trust with which they are unaffiliated
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('list_trust_roles',
                        expected_status=exceptions.Forbidden,
                        trust_id=trust_id)

    def test_identity_get_role_for_trust(self):
        # user can get roles for trust for which they are trustor
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        trust_id = self.client.create_trust(
            **self.trust(trustor=trustor_user_id,
                         project_id=project_id,
                         roles=[{'id': self.role_id}]))['trust']['id']
        self.addCleanup(self.client.delete_trust, trust_id=trust_id)
        self.do_request('show_trust_role',
                        trust_id=trust_id, role_id=self.role_id)

        # user can list roles for trust for which they are trustee
        trustee_user_id = self.persona.credentials.user_id
        trust_id = self.user_trust_client.create_trust(
            **self.trust(trustee=trustee_user_id))['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('show_trust_role',
                        trust_id=trust_id, role_id=self.roles[0]['id'])

        # user cannot list roles for trust with which they are unaffiliated
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('show_trust_role',
                        expected_status=exceptions.Forbidden,
                        trust_id=trust_id, role_id=self.role_id)

    def test_identity_delete_trust(self):
        # user can delete trust for which they are the trustor
        trustor_user_id = self.persona.credentials.user_id
        project_id = self.persona.credentials.project_id
        trust_id = self.client.create_trust(
            **self.trust(trustor=trustor_user_id,
                         project_id=project_id,
                         roles=[{'id': self.role_id}]))['trust']['id']
        self.do_request('delete_trust', expected_status=204, trust_id=trust_id)

        # user cannot delete trust for which they are the trustee
        trustee_user_id = self.persona.credentials.user_id
        trust_id = self.user_trust_client.create_trust(
            **self.trust(trustee=trustee_user_id))['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('delete_trust', expected_status=exceptions.Forbidden,
                        trust_id=trust_id)

        # user cannot delete trust with which they are unaffiliated
        trust_id = self.user_trust_client.create_trust(
            **self.trust())['trust']['id']
        self.addCleanup(self.user_trust_client.delete_trust, trust_id=trust_id)
        self.do_request('delete_trust', expected_status=exceptions.Forbidden,
                        trust_id=trust_id)


class ProjectMemberTests(ProjectManagerTests):

    credentials = ['project_member', 'system_admin']


class ProjectReaderTests(ProjectMemberTests):

    credentials = ['project_reader', 'system_admin']

    def setUp(self):
        super(ProjectReaderTests, self).setUp()
        self.role_id = self.reader_role_id



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib.common.utils import data_utils


def idp_ref(enabled=None, remote_ids=None):
    ref = {
        'description': data_utils.rand_uuid_hex(),
    }
    if enabled is not None:
        ref['enabled'] = enabled

    if remote_ids:
        ref['remote_ids'] = remote_ids

    return ref


def mapping_ref():
    rules = [{
        'local': [
            {
                'user': {'name': '{0}'}
            },
            {
                'group_ids': '{1}'
            }
        ],
        'remote': [
            {
                'type': 'openstack_username'
            },
            {
                'type': 'group_ids',
                'whitelist': ['abc', '123']
            }

        ]
    }]
    return {'rules': rules}


def sp_ref(enabled=None, relay_state_prefix=None):
    ref = {
        'auth_url': data_utils.rand_url(),
        'description': data_utils.rand_uuid_hex(),
        'sp_url': data_utils.rand_url(),
    }
    if enabled:
        ref['enabled'] = enabled

    if relay_state_prefix:
        ref['relay_state_prefix'] = relay_state_prefix

    return ref



# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tempest.api.identity import base
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from keystone_tempest_plugin.tests.rbac.v3 import base as rbac_base


class IdentityV3RbacAssignmentTest(rbac_base.IdentityV3RbacBaseTests,
                                   metaclass=abc.ABCMeta):

    @classmethod
    def setup_clients(cls):
        super(IdentityV3RbacAssignmentTest, cls).setup_clients()
        cls.persona = getattr(cls, 'os_%s' % cls.credentials[0])
        cls.client = cls.persona.role_assignments_client
        cls.admin_client = cls.os_system_admin

    @classmethod
    def resource_setup(cls):
        super(IdentityV3RbacAssignmentTest, cls).resource_setup()
        cls._setup_assignments()

    @classmethod
    def _setup_assignments(cls):
        cls.own_domain = cls.persona.credentials.domain_id
        cls.role_id = cls.admin_client.roles_v3_client.create_role(
            name=data_utils.rand_name('role'))['role']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.roles_v3_client.delete_role, cls.role_id)
        cls.user_in_domain = cls.admin_client.users_v3_client.create_user(
            name=data_utils.rand_name('user'),
            domain_id=cls.own_domain)['user']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.users_v3_client.delete_user,
            cls.user_in_domain)
        cls.group_in_domain = cls.admin_client.groups_client.create_group(
            name=data_utils.rand_name('group'),
            domain_id=cls.own_domain)['group']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.groups_client.delete_group,
            cls.group_in_domain)
        cls.project_in_domain = (
            cls.admin_client.projects_client.create_project(
                name=data_utils.rand_name('project'),
                domain_id=cls.own_domain)['project']['id'])
        cls.addClassResourceCleanup(
            cls.admin_client.projects_client.delete_project,
            cls.project_in_domain)
        cls.other_domain = cls.admin_client.domains_client.create_domain(
            name=data_utils.rand_name('domain'))['domain']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.domains_client.delete_domain,
            cls.other_domain)
        cls.addClassResourceCleanup(
            cls.admin_client.domains_client.update_domain,
            cls.other_domain,
            enabled=False)
        cls.user_other_domain = cls.admin_client.users_v3_client.create_user(
            name=data_utils.rand_name('user'),
            domain_id=cls.other_domain)['user']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.users_v3_client.delete_user,
            cls.user_other_domain)
        cls.group_other_domain = cls.admin_client.groups_client.create_group(
            name=data_utils.rand_name('group'),
            domain_id=cls.other_domain)['group']['id']
        cls.addClassResourceCleanup(
            cls.admin_client.groups_client.delete_group,
            cls.group_other_domain)
        cls.project_other_domain = (
            cls.admin_client.projects_client.create_project(
                name=data_utils.rand_name('project'),
                domain_id=cls.other_domain)['project']['id'])
        cls.addClassResourceCleanup(
            cls.admin_client.projects_client.delete_project,
            cls.project_other_domain)

        roles_client = cls.admin_client.roles_v3_client
        roles_client.create_user_role_on_project(
            cls.project_in_domain,
            cls.user_in_domain,
            cls.role_id)
        roles_client.create_user_role_on_project(
            cls.project_in_domain,
            cls.user_other_domain,
            cls.role_id)
        roles_client.create_user_role_on_project(
            cls.project_other_domain,
            cls.user_in_domain,
            cls.role_id)
        roles_client.create_user_role_on_project(
            cls.project_other_domain,
            cls.user_other_domain,
            cls.role_id)
        roles_client.create_user_role_on_domain(
            cls.own_domain,
            cls.user_in_domain,
            cls.role_id)
        roles_client.create_user_role_on_domain(
            cls.own_domain,
            cls.user_other_domain,
            cls.role_id)
        roles_client.create_user_role_on_domain(
            cls.other_domain,
            cls.user_in_domain,
            cls.role_id)
        roles_client.create_user_role_on_domain(
            cls.other_domain,
            cls.user_other_domain,
            cls.role_id)
        roles_client.create_user_role_on_system(
            cls.user_in_domain,
            cls.role_id)
        roles_client.create_user_role_on_system(
            cls.user_other_domain,
            cls.role_id)
        roles_client.create_group_role_on_project(
            cls.project_in_domain,
            cls.group_in_domain,
            cls.role_id)
        roles_client.create_group_role_on_project(
            cls.project_in_domain,
            cls.group_other_domain,
            cls.role_id)
        roles_client.create_group_role_on_project(
            cls.project_other_domain,
            cls.group_in_domain,
            cls.role_id)
        roles_client.create_group_role_on_project(
            cls.project_other_domain,
            cls.group_other_domain,
            cls.role_id)
        roles_client.create_group_role_on_domain(
            cls.own_domain,
            cls.group_in_domain,
            cls.role_id)
        roles_client.create_group_role_on_domain(
            cls.own_domain,
            cls.group_other_domain,
            cls.role_id)
        roles_client.create_group_role_on_domain(
            cls.other_domain,
            cls.group_in_domain,
            cls.role_id)
        roles_client.create_group_role_on_domain(
            cls.other_domain,
            cls.group_other_domain,
            cls.role_id)
        roles_client.create_group_role_on_system(
            cls.group_in_domain,
            cls.role_id)
        roles_client.create_group_role_on_system(
            cls.group_other_domain,
            cls.role_id)

        cls.assignments = [
            {
                'user_id': cls.user_in_domain,
                'project_id': cls.project_in_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_other_domain,
                'project_id': cls.project_in_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_in_domain,
                'project_id': cls.project_other_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_other_domain,
                'project_id': cls.project_other_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_in_domain,
                'domain_id': cls.own_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_other_domain,
                'domain_id': cls.own_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_in_domain,
                'domain_id': cls.other_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_other_domain,
                'domain_id': cls.other_domain,
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_in_domain,
                'system': 'all',
                'role_id': cls.role_id
            },
            {
                'user_id': cls.user_other_domain,
                'system': 'all',
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_in_domain,
                'project_id': cls.project_in_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_other_domain,
                'project_id': cls.project_in_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_in_domain,
                'project_id': cls.project_other_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_other_domain,
                'project_id': cls.project_other_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_in_domain,
                'domain_id': cls.own_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_other_domain,
                'domain_id': cls.own_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_in_domain,
                'domain_id': cls.other_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_other_domain,
                'domain_id': cls.other_domain,
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_in_domain,
                'system': 'all',
                'role_id': cls.role_id
            },
            {
                'group_id': cls.group_other_domain,
                'system': 'all',
                'role_id': cls.role_id
            },
        ]

    def _extract_role_assignments_from_response_body(self, r):
        # Condense the role assignment details into a set of key things we can
        # use in assertions.
        assignments = []
        for assignment in r['role_assignments']:
            a = {}
            if 'project' in assignment['scope']:
                a['project_id'] = assignment['scope']['project']['id']
            elif 'domain' in assignment['scope']:
                a['domain_id'] = assignment['scope']['domain']['id']
            elif 'system' in assignment['scope']:
                a['system'] = 'all'

            if 'user' in assignment:
                a['user_id'] = assignment['user']['id']
            elif 'group' in assignment:
                a['group_id'] = assignment['group']['id']

            a['role_id'] = assignment['role']['id']

            assignments.append(a)
        return assignments

    @abc.abstractmethod
    def test_identity_list_role_assignments(self):
        """Test identity:list_role_assignments policy.

        This test must check:
          * whether the persona can list all user and group assignments across
            the deployment
          * whether the persona can list all user and group assignments in
            a domain
          * whether the persona can list user and group assignments with names
          * whether the persona can filter user and group assignments by domain
          * whether the persona can filter user and group assignments by
            project in their own domain
          * whether the persona can filter user and group assignments by
            project in another domain
          * whether the persona can filter user and group assignments by system
          * whether the persona can filter user assignments by user in their
            own domain
          * whether the persona can filter user assignments by user in another
            domain
          * whether the persona can filter group assignments by group in their
            own domain
          * whether the persona can filter group assignments by group in
            another domain
          * whether the persona can filter role assignments by global role
          * whether the persona can filter assignments by project and role
          * whether the persona can filter assignments by domain and role
          * whether the persona can filter assignments by system and role
          * whether the persona can filter assignments by user and role
          * whether the persona can filter assignments by group and role
          * whether the persona can filter assignments by project and user
          * whether the persona can filter assignments by project and group
          * whether the persona can filter assignments by domain and user
          * whether the persona can filter assignments by domain and group

        """
        pass

    @abc.abstractmethod
    def test_identity_list_role_assignments_for_tree(self):
        """Test identity:list_role_assignments_for_tree policy.

        This test must check:
          * whether the persona can list role assignments for a subtree of a
            project in their own domain
          * whether the persona can list role assignments for a subtree of a
            project in another domain
          * whether the persona can list role assignments for a subtree of a
            project on which they have a role assignment (if applicable)
        """
        pass


class SystemAdminTests(IdentityV3RbacAssignmentTest, base.BaseIdentityTest):

    credentials = ['system_admin']

    def test_identity_list_role_assignments(self):
        # Listing all assignments with no filters should return all assignments
        resp = self.do_request('list_role_assignments')
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in self.assignments:
            self.assertIn(assignment, actual)

        # Listing all assignments with names
        query = {'include_names': True}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in self.assignments:
            self.assertIn(assignment, actual)

        # Filter assignments by own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.own_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.domain.id': self.own_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by other domain should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.other_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.domain.id': self.other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by project in own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.project.id': self.project_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by project in other domain should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_other_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.project.id': self.project_other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by system should succeed
        expected = [a for a in self.assignments if a.get('system') == 'all']
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.system': 'all'}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by user in own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('user_id') == self.user_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by user in other domain should succeed
        expected = [a for a in self.assignments
                    if a.get('user_id') == self.user_other_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by group in own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('group_id') == self.group_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by group in other domain should succeed
        expected = [a for a in self.assignments
                    if a.get('group_id') == self.group_other_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by global role should succeed
        expected = self.assignments
        query = {'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by project and role should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.project.id': self.project_in_domain,
                 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by domain and role should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.other_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.domain.id': self.other_domain, 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by system and role should succeed
        expected = [a for a in self.assignments if a.get('system') == 'all']
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.system': 'all', 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by user and role should succeed
        expected = [a for a in self.assignments
                    if a.get('user_id') == self.user_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain, 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by group and role should succeed
        expected = [a for a in self.assignments
                    if a.get('group_id') == self.group_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain, 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by project and user should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain
                    and a.get('user_id') == self.user_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain,
                 'scope.project.id': self.project_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by project and group should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain
                    and a.get('group_id') == self.group_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain,
                 'scope.project.id': self.project_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by domain and user should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.own_domain
                    and a.get('user_id') == self.user_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain,
                 'scope.domain.id': self.own_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by domain and group should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.own_domain
                    and a.get('group_id') == self.group_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain,
                 'scope.domain.id': self.own_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

    def test_identity_list_role_assignments_for_tree(self):
        # Should see subtree assignments for project in own domain
        subproject_id = self.admin_client.projects_client.create_project(
            name=data_utils.rand_name('project'),
            domain_id=self.own_domain,
            parent_id=self.project_in_domain)['project']['id']
        self.addCleanup(self.admin_client.projects_client.delete_project,
                        subproject_id)
        self.admin_client.roles_v3_client.create_user_role_on_project(
            subproject_id, self.user_in_domain, self.role_id)
        query = {'scope.project.id': self.project_in_domain,
                 'include_subtree': True}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        expected_assignment = {'user_id': self.user_in_domain,
                               'project_id': subproject_id,
                               'role_id': self.role_id}
        self.assertIn(expected_assignment, actual)

        # Should see subtree assignments for project in other domain
        subproject_id = self.admin_client.projects_client.create_project(
            name=data_utils.rand_name('project'),
            domain_id=self.other_domain,
            parent_id=self.project_other_domain)['project']['id']
        self.addCleanup(self.admin_client.projects_client.delete_project,
                        subproject_id)
        self.admin_client.roles_v3_client.create_user_role_on_project(
            subproject_id, self.user_in_domain, self.role_id)
        query = {'scope.project.id': self.project_other_domain,
                 'include_subtree': True}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        expected_assignment = {'user_id': self.user_in_domain,
                               'project_id': subproject_id,
                               'role_id': self.role_id}
        self.assertIn(expected_assignment, actual)


class SystemMemberTests(SystemAdminTests):

    credentials = ['system_member', 'system_admin']


class SystemReaderTests(SystemMemberTests):

    credentials = ['system_reader', 'system_admin']


class DomainAdminTests(IdentityV3RbacAssignmentTest, base.BaseIdentityTest):

    credentials = ['domain_admin', 'system_admin']

    def test_identity_list_role_assignments(self):
        # Listing all assignments with no filters should only return
        # assignments in own domain
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain
                    or a.get('domain_id') == self.own_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        resp = self.do_request('list_role_assignments')
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Listing all assignments with names and no filters should only return
        # assignments in own domain
        query = {'include_names': True}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.own_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.domain.id': self.own_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by other domain should be empty
        query = {'scope.domain.id': self.other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        self.assertEmpty(actual)

        # Filter assignments by project in own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.project.id': self.project_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by project in other domain should be empty
        query = {'scope.project.id': self.project_other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        self.assertEmpty(actual)

        # Filter assignments by system should be empty
        query = {'scope.system': 'all'}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        self.assertEmpty(actual)

        # Filter assignments by user in own domain should get assignments for
        # that user only for projects in own domain or for own domain itself
        expected = [a for a in self.assignments
                    if a.get('user_id') == self.user_in_domain
                    and (a.get('project_id') == self.project_in_domain
                         or a.get('domain_id') == self.own_domain)]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by user in other domain should still work but only
        # return assignments for projects in own domain or for own domain
        # itself
        expected = [a for a in self.assignments
                    if a.get('user_id') == self.user_other_domain
                    and (a.get('project_id') == self.project_in_domain
                         or a.get('domain_id') == self.own_domain)]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by group in own domain should succeed
        expected = [a for a in self.assignments
                    if a.get('group_id') == self.group_in_domain
                    and (a.get('project_id') == self.project_in_domain
                         or a.get('domain_id') == self.own_domain)]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by group in other domain should still work but
        # only return assignments for projects in own domain or for own domain
        # itself
        expected = [a for a in self.assignments
                    if a.get('group_id') == self.group_other_domain
                    and (a.get('project_id') == self.project_in_domain
                         or a.get('domain_id') == self.own_domain)]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_other_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by global role should only return role
        # assignments for own domain
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain
                    or a.get('domain_id') == self.own_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by project and role should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.project.id': self.project_in_domain,
                 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by domain and role should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.other_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'scope.domain.id': self.other_domain, 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by system and role should be empty
        query = {'scope.system': 'all', 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        self.assertEmpty(actual)

        # Filter assignments by user and role should should get assignments for
        # that user only for projects in own domain or for own domain itself
        expected = [a for a in self.assignments
                    if a.get('user_id') == self.user_in_domain
                    and (a.get('project_id') == self.project_in_domain
                         or a.get('domain_id') == self.own_domain)]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain, 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by group and role should get assignments for
        # that group only for projects in own domain or for own domain itself
        expected = [a for a in self.assignments
                    if a.get('group_id') == self.group_in_domain
                    and (a.get('project_id') == self.project_in_domain
                         or a.get('domain_id') == self.own_domain)]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain, 'role.id': self.role_id}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)
        # Reverse the check: only ephemeral tempest roles should be in the list
        for assignment in actual:
            self.assertIn(assignment, expected)

        # Filter assignments by project and user should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain
                    and a.get('user_id') == self.user_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain,
                 'scope.project.id': self.project_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by project and group should succeed
        expected = [a for a in self.assignments
                    if a.get('project_id') == self.project_in_domain
                    and a.get('group_id') == self.group_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain,
                 'scope.project.id': self.project_in_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by domain and user should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.own_domain
                    and a.get('user_id') == self.user_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'user.id': self.user_in_domain,
                 'scope.domain.id': self.own_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

        # Filter assignments by domain and group should succeed
        expected = [a for a in self.assignments
                    if a.get('domain_id') == self.own_domain
                    and a.get('group_id') == self.group_in_domain]
        not_expected = [a for a in self.assignments if a not in expected]
        query = {'group.id': self.group_in_domain,
                 'scope.domain.id': self.own_domain}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        for assignment in expected:
            self.assertIn(assignment, actual)
        for assignment in not_expected:
            self.assertNotIn(assignment, actual)

    def test_identity_list_role_assignments_for_tree(self):
        # Should see subtree assignments for project in own domain
        subproject_id = self.admin_client.projects_client.create_project(
            name=data_utils.rand_name('project'),
            domain_id=self.own_domain,
            parent_id=self.project_in_domain)['project']['id']
        self.addCleanup(self.admin_client.projects_client.delete_project,
                        subproject_id)
        self.admin_client.roles_v3_client.create_user_role_on_project(
            subproject_id, self.user_in_domain, self.role_id)
        query = {'scope.project.id': self.project_in_domain,
                 'include_subtree': True}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        expected_assignment = {'user_id': self.user_in_domain,
                               'project_id': subproject_id,
                               'role_id': self.role_id}
        self.assertIn(expected_assignment, actual)

        # Should not see subtree assignments for project in other domain
        query = {'scope.project.id': self.project_other_domain,
                 'include_subtree': True}
        self.do_request('list_role_assignments', **query)


class DomainMemberTests(DomainAdminTests):

    credentials = ['domain_member', 'system_admin']

    def test_identity_list_role_assignments_for_tree(self):
        # Should see subtree assignments for project in own domain
        subproject_id = self.admin_client.projects_client.create_project(
            name=data_utils.rand_name('project'),
            domain_id=self.own_domain,
            parent_id=self.project_in_domain)['project']['id']
        self.addCleanup(self.admin_client.projects_client.delete_project,
                        subproject_id)
        self.admin_client.roles_v3_client.create_user_role_on_project(
            subproject_id, self.user_in_domain, self.role_id)
        query = {'scope.project.id': self.project_in_domain,
                 'include_subtree': True}
        resp = self.do_request('list_role_assignments', **query)
        actual = self._extract_role_assignments_from_response_body(resp)
        expected_assignment = {'user_id': self.user_in_domain,
                               'project_id': subproject_id,
                               'role_id': self.role_id}
        self.assertIn(expected_assignment, actual)

        # Should not see subtree assignments for project in other domain
        query = {'scope.project.id': self.project_other_domain,
                 'include_subtree': True}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)


class DomainReaderTests(DomainMemberTests):

    credentials = ['domain_reader', 'system_admin']


class ProjectAdminTests(SystemAdminTests):

    credentials = ['project_admin', 'system_admin']


class ProjectManagerTests(IdentityV3RbacAssignmentTest, base.BaseIdentityTest):

    credentials = ['project_manager', 'system_admin']

    def test_identity_list_role_assignments(self):
        # Listing all assignments with no filters should fail
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden)

        # Listing all assignments with names and no filters should fail
        query = {'include_names': True}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by own domain should fail
        query = {'scope.domain.id': self.own_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by other domain should fail
        query = {'scope.domain.id': self.other_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by project in own domain should fail
        query = {'scope.project.id': self.project_in_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by project in other domain should fail
        query = {'scope.project.id': self.project_other_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by system should fail
        query = {'scope.system': 'all'}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by user in own domain should fail
        query = {'user.id': self.user_in_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by user in other domain should fail
        query = {'user.id': self.user_other_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by group in own domain should fail
        query = {'group.id': self.group_in_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by group in other domain should fail
        query = {'group.id': self.group_other_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by global role should fail
        query = {'role.id': self.role_id}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by project and role should fail
        query = {'scope.project.id': self.project_in_domain,
                 'role.id': self.role_id}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by domain and role should fail
        query = {'scope.domain.id': self.other_domain, 'role.id': self.role_id}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by system and role should fail
        query = {'scope.system': 'all', 'role.id': self.role_id}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by user and role should should fail
        query = {'user.id': self.user_in_domain, 'role.id': self.role_id}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by group and role should fail
        query = {'group.id': self.group_in_domain, 'role.id': self.role_id}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by project and user should fail
        query = {'user.id': self.user_in_domain,
                 'scope.project.id': self.project_in_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by project and group should fail
        query = {'group.id': self.group_in_domain,
                 'scope.project.id': self.project_in_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by domain and user should fail
        query = {'user.id': self.user_in_domain,
                 'scope.domain.id': self.own_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Filter assignments by domain and group should fail
        query = {'group.id': self.group_in_domain,
                 'scope.domain.id': self.own_domain}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

    def test_identity_list_role_assignments_for_tree(self):
        # Should not see subtree assignments for project in own domain
        query = {'scope.project.id': self.project_in_domain,
                 'include_subtree': True}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Should not see subtree assignments for project in other domain
        query = {'scope.project.id': self.project_other_domain,
                 'include_subtree': True}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)

        # Should not see subtree for own project
        own_project = self.persona.credentials.project_id
        query = {'scope.project.id': own_project,
                 'include_subtree': True}
        self.do_request('list_role_assignments',
                        expected_status=exceptions.Forbidden, **query)


class ProjectMemberTests(ProjectManagerTests):

    credentials = ['project_member', 'system_admin']


class ProjectReaderTests(ProjectMemberTests):

    credentials = ['project_reader', 'system_admin']



# Copyright 2022 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import identity
from keystoneauth1 import session as ks_session
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions
import testtools

from .keycloak import KeycloakClient
from keystone_tempest_plugin.tests import base

CONF = config.CONF


class TestOidcFederatedAuthentication(base.BaseIdentityTest):

    def _setup_settings(self):
        # Keycloak Settings
        self.idp_id = CONF.fed_scenario.idp_id
        self.idp_remote_ids = CONF.fed_scenario.idp_remote_ids
        self.idp_url = CONF.fed_scenario.idp_oidc_url
        self.idp_client_id = CONF.fed_scenario.idp_client_id
        self.idp_client_secret = CONF.fed_scenario.idp_client_secret
        self.idp_password = CONF.fed_scenario.idp_password
        self.idp_username = CONF.fed_scenario.idp_username

        self.protocol_id = CONF.fed_scenario.protocol_id
        self.keystone_v3_endpoint = CONF.identity.uri_v3

        # mapping settings
        self.mapping_remote_type = CONF.fed_scenario.mapping_remote_type
        self.mapping_user_name = CONF.fed_scenario.mapping_user_name
        self.mapping_group_name = CONF.fed_scenario.mapping_group_name
        self.mapping_group_domain_name = \
            CONF.fed_scenario.mapping_group_domain_name

        # custom CA certificate settings
        self.ca_certificates_file = CONF.identity.ca_certificates_file

    def _check_existing_protocol(self):
        try:
            self.idps_client.get_protocol_and_mapping(
                self.idp_id, self.protocol_id)
            return True
        except exceptions.NotFound:
            return False

    def _setup_mapping(self):
        self.mapping_id = data_utils.rand_uuid_hex()
        rules = [{
            'local': [
                {
                    'user': {'name': self.mapping_user_name}
                },
                {
                    'group': {
                        'domain': {'name': self.mapping_group_domain_name},
                        'name': self.mapping_group_name
                    }
                }
            ],
            'remote': [
                {
                    'type': self.mapping_remote_type
                }
            ]
        }]
        mapping_ref = {'rules': rules}
        self.mappings_client.create_mapping_rule(self.mapping_id, mapping_ref)
        self.addCleanup(
            self.mappings_client.delete_mapping_rule, self.mapping_id)

    def _setup_protocol(self):
        self.idps_client.add_protocol_and_mapping(
            self.idp_id, self.protocol_id, self.mapping_id)
        self.addCleanup(
            self.idps_client.delete_protocol_and_mapping,
            self.idp_id,
            self.protocol_id)

    def _setup_user(self, email=None):
        email = email if email else f'test-{uuid.uuid4().hex}@example.com'
        self.keycloak.create_user(email, 'Test', 'User')
        return email

    def _request_unscoped_token(self, user, password):
        auth = identity.v3.OidcPassword(
            auth_url=self.keystone_v3_endpoint,
            identity_provider=self.idp_id,
            protocol=self.protocol_id,
            client_id=self.idp_client_id,
            client_secret=self.idp_client_secret,
            access_token_endpoint=self.keycloak.token_endpoint,
            discovery_endpoint=self.keycloak.discovery_endpoint,
            username=user,
            password=password
        )
        s = ks_session.Session(auth, verify=self.ca_certificates_file)
        return s.get_auth_headers()

    def setUp(self):
        super(TestOidcFederatedAuthentication, self).setUp()
        self._setup_settings()

        # Setup mapping and protocol
        if not self._check_existing_protocol():
            self._setup_mapping()
            self._setup_protocol()

        self.keycloak = KeycloakClient(
            keycloak_url=self.idp_url,
            keycloak_username=self.idp_username,
            keycloak_password=self.idp_password,
            ca_certs_file=self.ca_certificates_file,
        )

        if CONF.fed_scenario.idp_test_user_name:
            self.test_user = CONF.fed_scenario.idp_test_user_name
            self.test_user_password = CONF.fed_scenario.idp_test_user_password
        else:
            self.test_user = self._setup_user()
            self.test_user_password = 'secret'

    @testtools.skipUnless(CONF.identity_feature_enabled.federation,
                          "Federated Identity feature not enabled")
    @testtools.skipUnless(CONF.identity_feature_enabled.external_idp,
                          "External identity provider is not available")
    @testtools.skipUnless(CONF.fed_scenario.protocol_id == 'openid',
                          "Protocol not openid")
    def test_request_unscoped_token(self):
        token = self._request_unscoped_token(self.test_user,
                                             self.test_user_password)
        self.assertNotEmpty(token)

    @testtools.skipUnless(CONF.identity_feature_enabled.federation,
                          "Federated Identity feature not enabled")
    @testtools.skipUnless(CONF.identity_feature_enabled.external_idp,
                          "External identity provider is not available")
    @testtools.skipUnless(CONF.fed_scenario.protocol_id == 'openid',
                          "Protocol not openid")
    def test_request_scoped_token(self):
        token = self._request_unscoped_token(self.test_user,
                                             self.test_user_password)
        token_id = token['X-Auth-Token']

        projects = self.auth_client.get_available_projects_scopes(
            self.keystone_v3_endpoint, token_id)['projects']
        self.assertNotEmpty(projects)

        # Get a scoped token to one of the listed projects
        self.tokens_client.auth(
            project_id=projects[0]['id'], token=token_id)

    def tearDown(self):
        super(TestOidcFederatedAuthentication, self).tearDown()
        if not CONF.fed_scenario.idp_test_user_name:
            self.keycloak.delete_user(self.test_user)



# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest import config

CONF = config.CONF


class IdentityV3RbacBaseTests(object):

    identity_version = 'v3'

    @classmethod
    def skip_checks(cls):
        super(IdentityV3RbacBaseTests, cls).skip_checks()
        if not CONF.identity_feature_enabled.enforce_scope:
            raise cls.skipException("enforce_scope is not enabled for "
                                    "keystone, skipping RBAC tests")

    def do_request(self, method, expected_status=200, client=None, **payload):
        if not client:
            client = self.client
        if isinstance(expected_status, type(Exception)):
            self.assertRaises(expected_status,
                              getattr(client, method),
                              **payload)
        else:
            response = getattr(client, method)(**payload)
            self.assertEqual(response.response.status, expected_status)
            return response



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from keystone_tempest_plugin.tests.api.identity.v3 import fixtures
from keystone_tempest_plugin.tests import base

DEFAULT_RELAY_STATE_PREFIX = 'ss:mem:'


class ServiceProvidersTest(base.BaseIdentityTest):

    def _assert_service_provider_attributes(self, sp, sp_id, sp_ref=None):
        self.assertIn('id', sp)
        self.assertEqual(sp_id, sp['id'])

        self.assertIn('auth_url', sp)
        self.assertIn('sp_url', sp)

        # Check the optional attributes have been set
        self.assertIn('description', sp)
        self.assertIn('enabled', sp)
        self.assertIn('relay_state_prefix', sp)

        if sp_ref:
            self.assertEqual(sp_ref['auth_url'], sp['auth_url'])
            self.assertEqual(sp_ref['sp_url'], sp['sp_url'])
            self.assertEqual(sp_ref['description'], sp['description'])

            if 'enabled' in sp_ref:
                self.assertEqual(sp_ref['enabled'], sp['enabled'])

            if 'relay_state_prefix' in sp_ref:
                self.assertEqual(
                    sp_ref['relay_state_prefix'], sp['relay_state_prefix'])

    def _create_sp(self, sp_id, sp_ref):
        sp = self.sps_client.create_service_provider(
            sp_id, **sp_ref)['service_provider']
        self.addCleanup(self.sps_client.delete_service_provider, sp_id)
        return sp

    @decorators.idempotent_id('6fae0971-5acb-4559-ba25-96f1fd7e5385')
    def test_service_provider_create(self):
        sp_id = data_utils.rand_uuid_hex()
        sp_ref = fixtures.sp_ref()
        sp = self._create_sp(sp_id, sp_ref)

        # The service provider is disabled by default
        sp_ref['enabled'] = False

        # The relay_state_prefix should have been set to the default value
        sp_ref['relay_state_prefix'] = DEFAULT_RELAY_STATE_PREFIX

        self._assert_service_provider_attributes(sp, sp_id, sp_ref)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('d9d7454c-50b7-4966-aedb-b9d520a41409')
    def test_service_provider_create_without_mandatory_attributes(self):
        sp_id = data_utils.rand_uuid_hex()
        self.assertRaises(
            lib_exc.BadRequest,
            self.sps_client.create_service_provider,
            sp_id)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('f77ed1c0-c428-44a7-9364-e8e4362c360a')
    def test_service_provider_create_with_bad_attributes(self):
        sp_id = data_utils.rand_uuid_hex()
        sp_ref = fixtures.sp_ref()

        # The auth_url must follow a URL regex
        sp_ref['auth_url'] = data_utils.rand_uuid_hex()
        self.assertRaises(
            lib_exc.BadRequest,
            self.sps_client.create_service_provider,
            sp_id,
            **sp_ref)

        sp_ref = fixtures.sp_ref()

        # The sp_url must follow a URL regex
        sp_ref['sp_url'] = data_utils.rand_uuid_hex()
        self.assertRaises(
            lib_exc.BadRequest,
            self.sps_client.create_service_provider,
            sp_id,
            **sp_ref)

    @decorators.idempotent_id('8550b419-f212-4e34-a8fa-7ff64f8a7fd3')
    def test_service_provider_create_with_enabled_true(self):
        sp_id = data_utils.rand_uuid_hex()
        sp_ref = fixtures.sp_ref(enabled=True)
        sp = self._create_sp(sp_id, sp_ref)

        self._assert_service_provider_attributes(sp, sp_id, sp_ref)

    @decorators.idempotent_id('0e319a14-1548-474e-a406-273c6b1c1f2d')
    def test_service_provider_create_with_relay_state_prefix(self):
        sp_id = data_utils.rand_uuid_hex()
        sp_ref = fixtures.sp_ref(
            enabled=True, relay_state_prefix=data_utils.rand_uuid_hex())
        sp = self._create_sp(sp_id, sp_ref)

        self._assert_service_provider_attributes(sp, sp_id, sp_ref)

    @decorators.idempotent_id('7df78c7a-9265-4b4f-9630-193b7f07d9eb')
    def test_service_provider_get(self):
        sp_id = data_utils.rand_uuid_hex()
        sp_create = self._create_sp(sp_id, fixtures.sp_ref())

        sp_get = self.sps_client.show_service_provider(sp_id)[
            'service_provider']

        self._assert_service_provider_attributes(sp_get, sp_id, sp_create)

    @decorators.idempotent_id('9237cea0-fbeb-4d64-8347-46c567e1d78f')
    def test_service_provider_list(self):
        sp_ids = []
        for _ in range(3):
            sp_id = data_utils.rand_uuid_hex()
            self._create_sp(sp_id, fixtures.sp_ref())
            sp_ids.append(sp_id)

        sp_list = self.sps_client.list_service_providers()['service_providers']
        fetched_ids = [fetched_sp['id'] for fetched_sp in sp_list]

        for sp_id in sp_ids:
            self.assertIn(sp_id, fetched_ids)

    @decorators.idempotent_id('bb68653f-fbba-4f20-ac1b-7b318a557366')
    def test_service_provider_update(self):
        sp_id = data_utils.rand_uuid_hex()
        sp = self._create_sp(sp_id, fixtures.sp_ref(enabled=True))

        # The service provider should be enabled
        self.assertTrue(sp['enabled'])

        sp = self.sps_client.update_service_provider(
            sp_id, enabled=False)['service_provider']

        # The service provider should be now disabled
        self.assertFalse(sp['enabled'])

        sp_get = self.sps_client.show_service_provider(sp_id)[
            'service_provider']
        self.assertFalse(sp_get['enabled'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('91ce1183-1a15-4598-ae5f-85cfa98a1c77')
    def test_service_provider_update_with_bad_attributes_fails(self):
        sp_id = data_utils.rand_uuid_hex()
        self._create_sp(sp_id, fixtures.sp_ref())

        # The auth_url must follow a URL regex
        self.assertRaises(
            lib_exc.BadRequest,
            self.sps_client.update_service_provider,
            sp_id,
            auth_url=data_utils.rand_uuid_hex())

        # The sp_url must follow a URL regex
        self.assertRaises(
            lib_exc.BadRequest,
            self.sps_client.update_service_provider,
            sp_id,
            auth_url=data_utils.rand_uuid_hex())

    @decorators.idempotent_id('7553579b-9a9e-45dd-9ada-70d906b516c0')
    def test_service_providers_in_token(self):
        # Create some enabled service providers
        enabled_sps = []
        for _ in range(2):
            sp_id = data_utils.rand_uuid_hex()
            self._create_sp(sp_id, fixtures.sp_ref(enabled=True))
            enabled_sps.append(sp_id)

        # Create some disabled service providers
        disabled_sps = []
        for _ in range(2):
            sp_id = data_utils.rand_uuid_hex()
            self._create_sp(sp_id, fixtures.sp_ref(enabled=False))
            disabled_sps.append(sp_id)

        sps_in_token_ids = [
            sp['id'] for sp in
            self.sps_client.get_service_providers_in_token()]

        for enabled_sp in enabled_sps:
            self.assertIn(enabled_sp, sps_in_token_ids)

        for disabled_sp in disabled_sps:
            self.assertNotIn(disabled_sp, sps_in_token_ids)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

import http.client
from tempest import config
from tempest.lib.common import rest_client


CONF = config.CONF

# We only use the identity catalog type
SERVICE_TYPE = 'identity'


class Identity(rest_client.RestClient):
    """Tempest REST client for keystone."""

    # Used by the superclass to build the correct URL paths
    api_version = 'v3'

    def __init__(self, auth_provider):
        super(Identity, self).__init__(
            auth_provider,
            SERVICE_TYPE,
            CONF.identity.region,
            endpoint_type=CONF.identity.v3_endpoint_type)


class Federation(Identity):
    """Tempest REST client for keystone's Federated Identity API."""

    subpath_prefix = 'OS-FEDERATION'
    subpath_suffix = None

    def _build_path(self, entity_id=None):
        subpath = '%s/%s' % (self.subpath_prefix, self.subpath_suffix)
        return '%s/%s' % (subpath, entity_id) if entity_id else subpath

    def _delete(self, entity_id, **kwargs):
        url = self._build_path(entity_id)
        resp, body = super(Federation, self).delete(url, **kwargs)
        self.expected_success(http.client.NO_CONTENT, resp.status)
        return rest_client.ResponseBody(resp, body)

    def _get(self, entity_id=None, **kwargs):
        url = self._build_path(entity_id)
        resp, body = super(Federation, self).get(url, **kwargs)
        self.expected_success(http.client.OK, resp.status)
        body = json.loads(body.decode('utf-8'))
        return rest_client.ResponseBody(resp, body)

    def _patch(self, entity_id, body, **kwargs):
        url = self._build_path(entity_id)
        resp, body = super(Federation, self).patch(url, body, **kwargs)
        self.expected_success(http.client.OK, resp.status)
        body = json.loads(body.decode('utf-8'))
        return rest_client.ResponseBody(resp, body)

    def _put(self, entity_id, body, **kwargs):
        url = self._build_path(entity_id)
        resp, body = super(Federation, self).put(url, body, **kwargs)
        self.expected_success(http.client.CREATED, resp.status)
        body = json.loads(body.decode('utf-8'))
        return rest_client.ResponseBody(resp, body)



# Copyright 2020 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tempest.api.identity import base
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from keystone_tempest_plugin.tests.rbac.v3 import base as rbac_base


class IdentityV3RbacUserTest(rbac_base.IdentityV3RbacBaseTests,
                             metaclass=abc.ABCMeta):

    @classmethod
    def setup_clients(cls):
        super(IdentityV3RbacUserTest, cls).setup_clients()
        cls.persona = getattr(cls, 'os_%s' % cls.credentials[0])
        cls.client = cls.persona.users_v3_client
        admin_client = cls.os_system_admin
        cls.admin_users_client = admin_client.users_v3_client
        cls.admin_domains_client = admin_client.domains_client

    def user(self):
        user = {}
        name = data_utils.rand_name('user')
        user['name'] = name
        user['description'] = name + 'description'
        user['email'] = name + '@testmail.tm'
        user['password'] = data_utils.rand_password()
        user['enabled'] = False
        return user

    @abc.abstractmethod
    def test_identity_create_user(self):
        """Test identity:create_user policy.

        This test must check:
          * whether the persona can create an arbitrary user
          * whether the persona can create a user in another domain
            (if applicable)
          * whether the persona can create a user in their own domain
            (if applicable)
        """
        pass

    @abc.abstractmethod
    def test_identity_get_user(self):
        """Test identity:get_user policy.

        This test must check:
          * whether the persona can get an arbitary user
          * whether the persona can get their own user
          * whether the persona can get a user in another domain
            (if applicable)
          * whether the persona can get a user in their own domain
            (if applicable)
          * whether the persona can get a user that does not exist
        """
        pass

    @abc.abstractmethod
    def test_identity_list_users(self):
        """Test identity:list_users policy.

        This test must check:
          * whether the persona can list all users
          * whether the result list is appropriately filtered to domain
            (if applicable)
        """
        pass

    @abc.abstractmethod
    def test_identity_update_user(self):
        """Test identity:update_users policy.

        This test must check:
          * whether the persona can update an arbitrary user
          * whether the persona can update a user in another domain
            (if applicable)
          * whether the persona can update a user in their own domain
            (if applicable)
          * whether the persona can update a user that does not exist
        """
        pass

    @abc.abstractmethod
    def test_identity_delete_user(self):
        """Test identity:delete_user policy.

        This test must check
          * whether the persona can delete an arbitrary user
          * whether the persona can delete a user in another domain
            (if applicable)
          * whether the persona can delete a user in their own domain
            (if applicable)
          * whether the persona can delete a user that does not exist
        """
        pass


class SystemAdminTests(IdentityV3RbacUserTest, base.BaseIdentityTest):

    credentials = ['system_admin']

    def test_identity_create_user(self):
        resp = self.do_request('create_user',
                               expected_status=201,
                               **self.user())
        self.addCleanup(self.admin_users_client.delete_user,
                        resp['user']['id'])

    def test_identity_get_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user can get arbitrary user
        resp = self.do_request('show_user', user_id=user['id'])
        self.assertEqual(resp['user']['id'], user['id'])
        # user can get own user
        user_id = self.persona.credentials.user_id
        resp = self.do_request('show_user', user_id=user_id)
        self.assertEqual(resp['user']['id'], user_id)
        # user gets a 404 for nonexistent user
        self.do_request('show_user', expected_status=exceptions.NotFound,
                        user_id='fakeuser')

    def test_identity_list_users(self):
        domain = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']
        user_create = self.user()
        # create user in default domain
        user1 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user1['id'])
        # create user in arbitrary domain
        user_create['domain_id'] = domain['id']
        user2 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user2['id'])
        resp = self.do_request('list_users')
        user_ids = set(u['id'] for u in resp['users'])
        # both users should be in the list
        self.assertIn(user1['id'], user_ids)
        self.assertIn(user2['id'], user_ids)

    def test_identity_update_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', **user_update)
        # user gets a 404 for nonexistent user
        user_update = {
            'user_id': 'fakeuser',
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.NotFound,
                        **user_update)

    def test_identity_delete_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.do_request('delete_user', expected_status=204, user_id=user['id'])
        # user gets a 404 for nonexistent user
        self.do_request('delete_user', expected_status=exceptions.NotFound,
                        user_id='fakeuser')


class SystemMemberTests(IdentityV3RbacUserTest, base.BaseIdentityTest):

    credentials = ['system_member', 'system_admin']

    def test_identity_create_user(self):
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **self.user())

    def test_identity_get_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user can get arbitrary user
        resp = self.do_request('show_user', user_id=user['id'])
        self.assertEqual(resp['user']['id'], user['id'])
        # user can get own user
        user_id = self.persona.credentials.user_id
        resp = self.do_request('show_user', user_id=user_id)
        self.assertEqual(resp['user']['id'], user_id)
        # user gets a 404 for nonexistent user
        self.do_request('show_user', expected_status=exceptions.NotFound,
                        user_id='fakeuser')

    def test_identity_list_users(self):
        domain = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']
        user_create = self.user()
        # create user in default domain
        user1 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user1['id'])
        # create user in arbitrary domain
        user_create['domain_id'] = domain['id']
        user2 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user2['id'])
        resp = self.do_request('list_users')
        user_ids = set(u['id'] for u in resp['users'])
        # both users should be in the list
        self.assertIn(user1['id'], user_ids)
        self.assertIn(user2['id'], user_ids)

    def test_identity_update_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)
        # user gets a 403 for nonexistent user
        user_update = {
            'user_id': 'fakeuser',
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)

    def test_identity_delete_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        # user gets a 403 for nonexistent user
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')


class SystemReaderTests(SystemMemberTests):

    credentials = ['system_reader', 'system_admin']


class DomainAdminTests(SystemAdminTests):

    credentials = ['domain_admin', 'system_admin']

    def test_identity_list_users(self):
        # the /v3/users api filters the response for domain-scoped
        # tokens to users only in the domain, so we can't re-use
        # the test from SystemAdminTests
        user_create = self.user()
        user_create['domain_id'] = self.persona.credentials.domain_id
        # create user in default domain
        user1 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user1['id'])
        resp = self.do_request('list_users')
        user_ids = set(u['id'] for u in resp['users'])
        self.assertIn(user1['id'], user_ids)


class DomainManagerTests(DomainAdminTests):

    credentials = ['domain_manager', 'system_admin']

    def setUp(self):
        super(DomainManagerTests, self).setUp()
        self.other_domain = self.admin_domains_client.create_domain(
            name=data_utils.rand_name())['domain']
        self.addCleanup(self.admin_domains_client.delete_domain,
                        self.other_domain['id'])
        self.addCleanup(self.admin_domains_client.update_domain,
                        domain_id=self.other_domain['id'], enabled=False)

    def test_identity_create_user(self):
        user_create = self.user()
        # cannot create user without domain specified
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **user_create)
        # cannot create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **user_create)
        # can create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        resp = self.do_request('create_user', expected_status=201,
                               **user_create)
        self.addCleanup(self.admin_users_client.delete_user,
                        resp['user']['id'])

    def test_identity_get_user(self):
        user_create = self.user()
        user_create['domain_id'] = self.other_domain['id']
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user cannot get user in other domain
        self.do_request('show_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        user_create['domain_id'] = self.persona.credentials.domain_id
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user can get user in own domain
        resp = self.do_request('show_user', user_id=user['id'])
        self.assertEqual(resp['user']['id'], user['id'])
        # user can get own user
        user_id = self.persona.credentials.user_id
        resp = self.do_request('show_user', user_id=user_id)
        self.assertEqual(resp['user']['id'], user_id)
        # user gets a 403 for nonexistent user
        self.do_request('show_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')

    def test_identity_list_users(self):
        user_create = self.user()
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        user1 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user1['id'])
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        user2 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user2['id'])
        resp = self.do_request('list_users')
        user_ids = set(u['id'] for u in resp['users'])
        self.assertNotIn(user1['id'], user_ids)
        self.assertIn(user2['id'], user_ids)

    def test_identity_update_user(self):
        user_create = self.user()
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user can update user in own domain
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', **user_update)
        # user gets a 403 for nonexistent user
        user_update = {
            'user_id': 'fakeuser',
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)

    def test_identity_delete_user(self):
        user_create = self.user()
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user cannot delete user in other domain
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        user = self.admin_users_client.create_user(**user_create)['user']
        # user can delete user in own domain
        self.do_request('delete_user', expected_status=204, user_id=user['id'])
        # user gets a 403 for nonexistent user
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')


class DomainMemberTests(DomainManagerTests):

    credentials = ['domain_member', 'system_admin']

    def test_identity_create_user(self):
        user_create = self.user()
        # create user without domain specified
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **user_create)
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **user_create)
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **user_create)

    def test_identity_get_user(self):
        user_create = self.user()
        user_create['domain_id'] = self.other_domain['id']
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user cannot get user in other domain
        self.do_request('show_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        user_create['domain_id'] = self.persona.credentials.domain_id
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user can get user in own domain
        resp = self.do_request('show_user', user_id=user['id'])
        self.assertEqual(resp['user']['id'], user['id'])
        # user can get own user
        user_id = self.persona.credentials.user_id
        resp = self.do_request('show_user', user_id=user_id)
        self.assertEqual(resp['user']['id'], user_id)
        # user gets a 403 for nonexistent user
        self.do_request('show_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')

    def test_identity_list_users(self):
        user_create = self.user()
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        user1 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user1['id'])
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        user2 = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user2['id'])
        resp = self.do_request('list_users')
        user_ids = set(u['id'] for u in resp['users'])
        self.assertNotIn(user1['id'], user_ids)
        self.assertIn(user2['id'], user_ids)

    def test_identity_update_user(self):
        user_create = self.user()
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)
        # user gets a 403 for nonexistent user
        user_update = {
            'user_id': 'fakeuser',
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)

    def test_identity_delete_user(self):
        user_create = self.user()
        # create user in other domain
        user_create['domain_id'] = self.other_domain['id']
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        # create user in own domain
        user_create['domain_id'] = self.persona.credentials.domain_id
        user = self.admin_users_client.create_user(**user_create)['user']
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        # user gets a 403 for nonexistent user
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')


class DomainReaderTests(DomainMemberTests):

    credentials = ['domain_reader', 'system_admin']


class ProjectAdminTests(SystemAdminTests):

    credentials = ['project_admin', 'system_admin']


class ProjectManagerTests(IdentityV3RbacUserTest, base.BaseIdentityTest):

    credentials = ['project_manager', 'system_admin']

    def test_identity_create_user(self):
        self.do_request('create_user', expected_status=exceptions.Forbidden,
                        **self.user())

    def test_identity_get_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        # user cannot get arbitrary user
        self.do_request('show_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        # user can get own user
        user_id = self.persona.credentials.user_id
        resp = self.do_request('show_user', user_id=user_id)
        self.assertEqual(resp['user']['id'], user_id)
        # user gets a 403 for nonexistent user
        self.do_request('show_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')

    def test_identity_list_users(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        self.do_request('list_users', expected_status=exceptions.Forbidden)

    def test_identity_update_user(self):
        user_create = self.user()
        user = self.admin_users_client.create_user(**user_create)['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        user_update = {
            'user_id': user['id'],
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)
        # user gets a 403 for nonexistent user
        user_update = {
            'user_id': 'fakeuser',
            'description': data_utils.arbitrary_string()
        }
        self.do_request('update_user', expected_status=exceptions.Forbidden,
                        **user_update)

    def test_identity_delete_user(self):
        user = self.admin_users_client.create_user(**self.user())['user']
        self.addCleanup(self.admin_users_client.delete_user, user['id'])
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id=user['id'])
        # user gets a 403 for nonexistent user
        self.do_request('delete_user', expected_status=exceptions.Forbidden,
                        user_id='fakeuser')


class ProjectMemberTests(ProjectManagerTests):

    credentials = ['project_member', 'system_admin']


class ProjectReaderTests(ProjectMemberTests):

    credentials = ['project_reader', 'system_admin']
